/** \file     TDecEM_JEG.cpp
    \brief    EM_JEG decoder modification class to consider data loss while decoding frames.

 *
 * Decoder modifications to consider data loss while decoding frames. Source code compatible with HM 12.1
 * Author: Enrico Masala <masala@polito.it>, Politecnico di Torino, Feb 9, 2014
 *
 * Disclaimer:
 * This software is meant to be a prototype only, it is not optimized or fitted to any purposes.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * 
 *
 * Citing this work:
 * We kindly ask people using this code in their research activity
 * to cite the research papers explaining the logic behind this code.
 * More information at: http://media.polito.it/jeg

*/


#include "TDecEM_JEG.h"
#include "TDecGop.h"
#include "TDecCAVLC.h"
#include "TDecSbac.h"
#include "TDecBinCoder.h"
#include "TDecBinCoderCABAC.h"
#include "libmd5/MD5.h"
#include "TLibCommon/SEI.h"

#include <time.h>

#define EM_JEG_INT_MIN (-2147483647 - 1)
#define EM_JEG_INT_MAX (2147483647)


#ifdef WIN32
#define strdup _strdup
#endif


using namespace std;

//! \ingroup TLibDecoder
//! \{
// ====================================================================================================================
// Constructor / destructor / initialization / destroy
// ====================================================================================================================

TDecEM_JEG::TDecEM_JEG()
{
  m_pchEM_JEGLossFile = NULL;
  m_pchEM_JEGProcessedFile = NULL;
  m_pchEM_JEGNALUTypeFile = NULL;
  filedes.loss = NULL;
  filedes.processed_bitstream = NULL;
  //nLossEvents = 0;
  nSlices = 0;
  //storedPicYuv = NULL;
  //firstPictureFlag = 1;
  
  current_slice_info = NULL;

  EM_JEG_debug_print = 0;
  EM_JEG_debug_dpb_poc = NULL;

  EM_JEG_dpb = NULL;
  EM_JEG_dpb_max = 0;
  EM_JEG_dpb_size = 0;

  EM_JEG_frame_num = 0;
  EM_JEG_max_frame_num = 0;
  EM_JEG_flag_lossfirst = 0;
  EM_JEG_nalu = NULL;
  EM_JEG_nalu_isLost = 0;

  EM_JEG_oldPoc = -1;  // First POC is 0, so first time the comparison is done a new frame is detected
  EM_JEG_doShift = 0;

  m_cpTComEM_JEG = NULL;
}

TDecEM_JEG::~TDecEM_JEG()
{
  
}

void TDecEM_JEG::create()
{
  current_slice_info = (EM_JEG_slice_info_t *)malloc(sizeof(EM_JEG_slice_info_t));
  if (!current_slice_info) {
    printf("Cannot allocate a EM_JEG_slice_info_t element\n"); exit(1);
  }
  current_slice_info->cu = (EM_JEG_cu_info_t *)malloc(sizeof(EM_JEG_cu_info_t)*MAX_CU_NUM);
  if (!current_slice_info->cu) {
    printf("Cannot allocate a EM_JEG_cu_info_t element\n"); exit(1);
  }
  current_slice_info->num_cu=0;
  
  EM_JEG_nalu = (EM_JEG_nalu_t *)malloc(sizeof(EM_JEG_nalu_t));
  if (!EM_JEG_nalu) {
    printf("Cannot allocate a EM_JEG_nalu_t element\n"); exit(1);
  }
  EM_JEG_nalu->buf = (uint8_t *)malloc(sizeof(uint8_t)*MAX_NALU_SIZE);
  if (!EM_JEG_nalu->buf) {
    printf("Cannot allocate a EM_JEG_nalu->buf buffer\n"); exit(1);
  }
  EM_JEG_nalu->len = 0;
  
}


void TDecEM_JEG::destroy()
{
  if (filedes.loss)
    EM_JEG_close_file(filedes.loss);
  if (filedes.processed_bitstream)
    EM_JEG_close_file(filedes.processed_bitstream);
  if (filedes.nalu_type)
    EM_JEG_close_file(filedes.nalu_type);
  //if (storedPicYuv) {
  //  storedPicYuv->destroy();
  //}
}

void TDecEM_JEG::init(TComEM_JEG *cpTComEM_JEG)
{
  m_cpTComEM_JEG = cpTComEM_JEG;
  m_cpTComEM_JEG->init();
}


void TDecEM_JEG::setEM_JEGLossFile( char *EM_JEGLossFile )
{
  if (EM_JEGLossFile) {
    m_pchEM_JEGLossFile = strdup(EM_JEGLossFile);
  }
  if (m_pchEM_JEGLossFile) {
    printf("EM_JEG: Info: LossFile specified, name: %s\n", m_pchEM_JEGLossFile);
    EM_JEG_open_file_loss();
  } else {
    printf("EM_JEG: Info: LossFile NOT specified\n");
  }

}

void TDecEM_JEG::setEM_JEGProcessedFile( char *EM_JEGProcessedFile )
{
  if (EM_JEGProcessedFile) {
    m_pchEM_JEGProcessedFile = strdup(EM_JEGProcessedFile);
  }
  if (m_pchEM_JEGProcessedFile) {
    printf("EM_JEG: Info: ProcessedFile specified, name: %s\n", m_pchEM_JEGProcessedFile);
    EM_JEG_open_file_processed_bitstream();
  } else {
    //printf("EM_JEG: Info: ProcessedFile NOT specified\n");
  }

}

void TDecEM_JEG::setEM_JEGNALUTypeFile( char *EM_JEGNALUTypeFile )
{
  if (EM_JEGNALUTypeFile) {
    m_pchEM_JEGNALUTypeFile = strdup(EM_JEGNALUTypeFile);
  }
  if (m_pchEM_JEGNALUTypeFile) {
    printf("EM_JEG: Info: NALUTypeFile specified, name: %s\n", m_pchEM_JEGNALUTypeFile);
    EM_JEG_open_file_nalutype();
  } else {
    //printf("EM_JEG: Info: NALUTypeFile NOT specified\n");
  }

}


void TDecEM_JEG::setEM_JEGParameterSetManager( ParameterSetManager *parameterSetManager ) {
  //m_pchEM_JEGparameterSetManager = parameterSetManager;  
}




// ====================================================================================================================
// Private member functions
// ====================================================================================================================


void TDecEM_JEG::EM_JEG_init_buffers(TComPic *pcPic) {
	//DecodedPicList
	//int tlayers = pcPic->getSlice(0)->getVPS()->getMaxTLayers();
	//EM_JEG_dpb_max = pcPic->getSlice(0)->getVPS()->getMaxDecPicBuffering(0);
	//int t;
	//for (t=1; t<tlayers; t++) {
	//	int v = pcPic->getSlice(0)->getVPS()->getMaxDecPicBuffering(t);
	//	if (v > EM_JEG_dpb_max) {
	//		EM_JEG_dpb_max = v;
	//	}
	//}
	EM_JEG_dpb_max = MAX_NUM_REF_PICS;  // Easier

	int width = pcPic->getSlice(0)->getSPS()->getPicWidthInLumaSamples();
	int height = pcPic->getSlice(0)->getSPS()->getPicHeightInLumaSamples();
	int maxCUWidth = pcPic->getSlice(0)->getSPS()->getMaxCUWidth();
	int maxCUHeight = pcPic->getSlice(0)->getSPS()->getMaxCUHeight();
	int maxCUDepth = pcPic->getSlice(0)->getSPS()->getMaxCUDepth();

	//printf("DEBUG EM_JEG: EM_JEG_dpb_max=%d\n", EM_JEG_dpb_max);

	EM_JEG_dpb = (TComPicYuv **)malloc(sizeof(TComPicYuv *)*EM_JEG_dpb_max);
	if (!EM_JEG_dpb) { printf("ERROR: cannot allocate EM_JEG_dpb\n"); exit(1); }
	EM_JEG_dpb_info = (EM_JEG_dpb_info_t *)malloc(sizeof(EM_JEG_dpb_info_t)*EM_JEG_dpb_max);
	if (!EM_JEG_dpb_info) { printf("ERROR: cannot allocate EM_JEG_dpb_info\n"); exit(1); }
	EM_JEG_debug_dpb_poc = (int *)malloc(sizeof(int)*EM_JEG_dpb_max);
	if (!EM_JEG_debug_dpb_poc) { printf("ERROR: cannot allocate EM_JEG_debug_dpb_poc\n"); exit(1); }
	int i;

	for (i=0; i<EM_JEG_dpb_max; i++) {
		EM_JEG_dpb[i] = new TComPicYuv;
		if (!EM_JEG_dpb[i]) { printf("ERROR: cannot allocate EM_JEG_dpb[%d]\n",i); exit(1); }
		EM_JEG_dpb[i]->create(width, height, maxCUWidth, maxCUHeight, maxCUDepth);
		EM_JEG_dpb_info[i].poc=EM_JEG_INT_MIN;
		EM_JEG_dpb_info[i].valid=0;
		EM_JEG_dpb_info[i].type=-1;
	}

	//EM_JEG_allocate_storedPicYuv(pcPic);
}

//void TDecEM_JEG::EM_JEG_allocate_storedPicYuv(TComPic *pcPic) {
//	// This is called ONLY when width and height etc. have been set
//
//	storedPicYuv = new TComPicYuv();
//	int picWidth = pcPic->getSlice(0)->getSPS()->getPicWidthInLumaSamples();
//	int picHeight = pcPic->getSlice(0)->getSPS()->getPicHeightInLumaSamples();
//	storedPicYuv->create(picWidth, picHeight, g_uiMaxCUWidth, g_uiMaxCUHeight, g_uiMaxCUDepth);
//
//}

void TDecEM_JEG::EM_JEG_print_BitstreamData(uint8_t *buf, int len) {
	int i;
	int ini=8;
	int end=8;
	if (len>ini+end) {
		for (i=0; i<ini; i++) {
			printf("%02x ",(int)(unsigned char)buf[i]);
		}
		printf("... ");
		for (i=len-end; i<len; i++) {
			printf("%02x ",(int)(unsigned char)buf[i]);
		}
	} else {
		for (i=0; i<len; i++) {
			printf("%02x ",(int)(unsigned char)buf[i]);
		}
	}
}

//void TDecEM_JEG::EM_JEG_write_BitstreamData_if_needed(EM_JEG_nalu_t *nalu) 
void TDecEM_JEG::EM_JEG_private_write_BitstreamData_if_needed(void) {
	EM_JEG_nalu_t *nalu = EM_JEG_nalu;
	//printf("DEBUG EM_JEG writing NALU with nalu->len = %d : %s \n", nalu->len, EM_JEG_nalu_isLost?"NO":"YES");
	if (! EM_JEG_nalu_isLost) {
		if (EM_JEG_nalu->len > 0 ) { // There is a previous NALU
			//printf("Writing data: %5d ",nalu->len); EM_JEG_print_BitstreamData(nalu->buf, nalu->len); printf("\n");

			int ret = fwrite(nalu->buf, nalu->len, 1, filedes.processed_bitstream);
			if (ret!=1) {
				printf("ERROR: cannot write NALU in processed_bistream\n");
				exit(1);
			}
			if (filedes.nalu_type) {
				fprintf(filedes.nalu_type, "%s\n", TDecEM_JEG::EM_JEG_get_NALU_name(nalu->type));
			}
		}
	} else {
		EM_JEG_nalu_isLost = 0; // Reset the flag, previous NALU has not been written
	}
}

void TDecEM_JEG::EM_JEG_write_BitstreamData_if_needed(void) {
	if (filedes.processed_bitstream) {
		EM_JEG_private_write_BitstreamData_if_needed();
	}
}


void TDecEM_JEG::EM_JEG_save_Annexb_NALU(uint8_t *data, int len, enum NalUnitType naluType) {
	if (filedes.processed_bitstream) {
		// Done separately from the save operation
		//EM_JEG_write_BitstreamData_if_needed(EM_JEG_nalu);

		// Now save current NALU, later it will be determined if this is considered lost or not
		if (len>0) {
			if (len>MAX_NALU_SIZE) {
				printf("ERROR: len>MAX_NALU_SIZE (%d): increase MAX_NALU_SIZE and recompile\n",MAX_NALU_SIZE);
				exit(1);
			}
			::memcpy(EM_JEG_nalu->buf, data, len);
		}
		EM_JEG_nalu->len = len;
		EM_JEG_nalu->type = naluType;
	}
}


void TDecEM_JEG::EM_JEG_saveNALU(vector<uint8_t>& nalUnit, enum NalUnitType naluType) {
	//vector<uint8_t>& nalUnit;
	int len = nalUnit.size();
	uint8_t *data = &nalUnit[0];
	//printf("EM_JEG DEBUG: nalUnit.size() = %d\n", len);
	// pre-alloc data
	//memcpy(data, &nalUnit[0], nalUnit.size());
	EM_JEG_save_Annexb_NALU(data, len, naluType);
}


#define MAXLINE 100000
void TDecEM_JEG::EM_JEG_open_file_loss(void) {
	FILE *fp = fopen(m_pchEM_JEGLossFile, "rt");	
	if (fp == NULL) {
		printf("EM_JEG: Error opening loss file %s\n", m_pchEM_JEGLossFile);
		exit(1);
	}
	printf("EM_JEG: File %s opened!\n", m_pchEM_JEGLossFile);
	filedes.loss=fp;
}


void TDecEM_JEG::EM_JEG_open_file_processed_bitstream(void) {
	FILE *fp = fopen(m_pchEM_JEGProcessedFile, "wb");	
	if (fp == NULL) {
		printf("EM_JEG: Error opening processed bitstream file %s\n", m_pchEM_JEGProcessedFile);
		exit(1);
	}
	printf("EM_JEG: File %s opened!\n", m_pchEM_JEGProcessedFile);
	filedes.processed_bitstream=fp;
}


void TDecEM_JEG::EM_JEG_open_file_nalutype(void) {
	FILE *fp = fopen(m_pchEM_JEGNALUTypeFile, "wt");	
	if (fp == NULL) {
		printf("EM_JEG: Error opening NALU type file %s\n", m_pchEM_JEGNALUTypeFile);
		exit(1);
	}
	printf("EM_JEG: File %s opened!\n", m_pchEM_JEGNALUTypeFile);
	filedes.nalu_type=fp;
}


void TDecEM_JEG::EM_JEG_close_file(FILE *fp) {
	if (fp) {
		fclose(fp);
	}
}

#define MAX_BUF_EVENT 1024

int TDecEM_JEG::EM_JEG_get_loss_event(void) {
	int val;
	char buf[MAX_BUF_EVENT];

	if (fgets(buf, MAX_BUF_EVENT, filedes.loss) == NULL) {
		printf("EM_JEG: Error: Premature end on file of loss events!\n");
		exit(1);
	}
	val = buf[0] - '0';

	if (val != 0 && val != 1) {
		printf("EM_JEG: Error: event must be 0 or 1! (Read: 0x%02x = '%c' = %3d)\n", val, (char)val, val);
		exit(1);
	}
	return val;
}



void TDecEM_JEG::EM_JEG_apply_cu_concealment(TComPic *pcPic, TComPicYuv *pcYuvRef, int cu_address) {

	int i,j;

	TComPicYuv* pcYuv = pcPic->getPicYuvRec();

	// Get buf Y
	Pel* pY = pcYuv->getBufY();
	int leftOffsetY = pcYuv->getLumaMargin(); // Offset needed to discard the left and right margins (used for filtering etc.)
	int rowOffsetY = pcYuv->getStride(); // Offset needed to go to a new row
	pY = pY; // Avoid compiler warning due to printf commented
	leftOffsetY = leftOffsetY; // Avoid compiler warning due to printf commented
	rowOffsetY = rowOffsetY; // Avoid compiler warning due to printf commented

	int CUWidth = pcPic->getSlice(0)->getSPS()->getMaxCUWidth();
	int CUHeight = pcPic->getSlice(0)->getSPS()->getMaxCUHeight();
	int ImgWidth = pcPic->getSlice(0)->getSPS()->getPicWidthInLumaSamples();
	int ImgHeight = pcPic->getSlice(0)->getSPS()->getPicHeightInLumaSamples();

	//printf("EM_JEG: Info: leftOffsetY %d rowOffsetY %d CUwidth %d CUheight %d pY %p LumaMargin %d  CUaddress %d\n", leftOffsetY, rowOffsetY, CUWidth, CUHeight, pY, pcYuv->getLumaMargin(), cu_address);
	
	//int ImgWidthCU = (ImgWidth%CUWidth)? ImgWidth/CUWidth+1 : ImgWidth/CUWidth;
	//int ImgHeightCU = (ImgHeight%CUHeight)? ImgHeight/CUHeight+1 : ImgHeight/CUHeight;
	//int CUAddrCol = cu_address % ImgWidthCU;
	//int CUAddrRow = cu_address / ImgWidthCU;

	// Verify that writing outside the image is always covered by the margin (should be so, not sure...) to simplify code later
	if (pcYuv->getLumaMargin() < ImgWidth%CUWidth  ||  pcYuv->getLumaMargin() < ImgHeight%CUHeight) {
		fprintf(stderr,"EM_JEG: ERROR: internal error: assumption that margin is >= part of CU that is out of the image is not true!!! Some code has to be rewritten by checking borders!\n"); exit(1);
	}


	if (pcYuvRef!=NULL) {
		m_cpTComEM_JEG->EM_JEG_setCUConcealed(cu_address);

		// No border check since it should be covered by the margin
		// Y
		for (j=0; j<CUHeight; j++) {
			//Pel* pYImgCurr = pcYuv->getBufY() + pcYuv->EM_JEG_getCUOffsetY(cu_address) + j*pcYuv->getStride();
			int margin = pcYuv->getLumaMargin();
			int marginRef = pcYuvRef->getLumaMargin();
			Pel* pYImgCurr = pcYuv->getBufY() + margin + pcYuv->EM_JEG_getCUOffsetY(cu_address) + (j+margin)*pcYuv->getStride();
			Pel* pYImgRef = pcYuvRef->getBufY() + marginRef + pcYuvRef->EM_JEG_getCUOffsetY(cu_address) + (j+marginRef)*pcYuvRef->getStride();
			for (i=0; i<CUWidth; i++) {
				*pYImgCurr = *pYImgRef;
				pYImgCurr++;
				pYImgRef++;
			}
		}
		// U,V
		for (j=0; j<CUHeight/2; j++) {
			int margin = pcYuv->getChromaMargin();
			int marginRef = pcYuvRef->getChromaMargin();
			Pel* pUImgCurr = pcYuv->getBufU() + margin + pcYuv->EM_JEG_getCUOffsetC(cu_address) + (j+margin)*pcYuv->getCStride();
			Pel* pVImgCurr = pcYuv->getBufV() + margin + pcYuv->EM_JEG_getCUOffsetC(cu_address) + (j+margin)*pcYuv->getCStride();
			Pel* pUImgRef = pcYuvRef->getBufU() + marginRef + pcYuvRef->EM_JEG_getCUOffsetC(cu_address) + (j+marginRef)*pcYuvRef->getCStride();
			Pel* pVImgRef = pcYuvRef->getBufV() + marginRef + pcYuvRef->EM_JEG_getCUOffsetC(cu_address) + (j+marginRef)*pcYuvRef->getCStride();
			for (i=0; i<CUWidth/2; i++) {
				*pUImgCurr = *pUImgRef;
				pUImgCurr++;
				pUImgRef++;
				*pVImgCurr = *pVImgRef;
				pVImgCurr++;
				pVImgRef++;
			}
		}

	} else {
		if (m_pchEM_JEGLossFirst) {
			// In case of first frame
			// Assuming concealment with value 128
			m_cpTComEM_JEG->EM_JEG_setCUConcealed(cu_address);

			// Y
			for (j=0; j<CUHeight; j++) {
				//Pel* pYImgCurr = pcYuv->getBufY() + pcYuv->EM_JEG_getCUOffsetY(cu_address) + j*pcYuv->getStride();
				int margin = pcYuv->getLumaMargin();
				Pel* pYImgCurr = pcYuv->getBufY() + margin + pcYuv->EM_JEG_getCUOffsetY(cu_address) + (j+margin)*pcYuv->getStride();
				for (i=0; i<CUWidth; i++) {
					*pYImgCurr = 0x80; // 0x80 grey, 0x00 for black
					pYImgCurr++;
				}
			}
			// U,V
			for (j=0; j<CUHeight/2; j++) {
				int margin = pcYuv->getChromaMargin();
				Pel* pUImgCurr = pcYuv->getBufU() + margin + pcYuv->EM_JEG_getCUOffsetC(cu_address) + (j+margin)*pcYuv->getCStride();
				Pel* pVImgCurr = pcYuv->getBufV() + margin + pcYuv->EM_JEG_getCUOffsetC(cu_address) + (j+margin)*pcYuv->getCStride();
				for (i=0; i<CUWidth/2; i++) {
					*pUImgCurr = 0x80;
					pUImgCurr++;
					*pVImgCurr = 0x80;
					pVImgCurr++;
				}
			}
		}


//		// TODO FIXME: To finish: use the storedPicYuv if there is need of a previous frame which is not in the list, or the list if there is. (finish searching in the list...)
//
//		int minim = 1000000;
//		int chosen_poc = -1;
//		TComPic* p = NULL;
//		// Code adapted from TDecTop::xCreateLostPicture
//		TComList<TComPic*>::iterator iterPic = m_cListPic->begin();
//		printf("EM_JEG TComList size %ld POCs: ",  m_cListPic->size() );
//		while ( iterPic != m_cListPic->end()) {
//			TComPic *rpcPic = *(iterPic++);
//			int rpcPicPOC = rpcPic->getPicSym()->getSlice(0)->getPOC();
//			printf("%d ", rpcPicPOC);
//			int diff = poc_num__to_be_concealed - rpcPicPOC;
//			if (diff > 0 && diff < minim) {
//				minim = diff;
//				chosen_poc = rpcPicPOC;
//				p = rpcPic;
//			}
//			//if(abs(rpcPic->getPicSym()->getSlice(0)->getPOC() -iLostPoc)<closestPoc&&abs(rpcPic->getPicSym()->getSlice(0)->getPOC() -iLostPoc)!=0&&rpcPic->getPicSym()->getSlice(0)->getPOC()!=m_apcSlicePilot->getPOC()) {
//			//	closestPoc=abs(rpcPic->getPicSym()->getSlice(0)->getPOC() -iLostPoc);
//			//}
//		}
//		printf(" --- storedPicYuvPOC %d\n", storedPicYuvPOC);
//
//		TComPicYuv* pcYuvRef = NULL;
//		if (p) { // if !p it means that nothing is found in the past
//			pcYuvRef = p->getPicYuvRec();
//		}
//		int diff = poc_num__to_be_concealed - storedPicYuvPOC;
//		if (storedPicYuvPOC >= 0 && diff > 0 && diff < minim) {
//			minim = poc_num__to_be_concealed - storedPicYuvPOC;
//			chosen_poc = storedPicYuvPOC;
//			pcYuvRef = storedPicYuv; //->getPicYuvRec();
//		}
//		printf("EM_JEG: Info: POC chosen for concealment: %d\n", chosen_poc);
//		if (!pcYuvRef) {
//			fprintf(stderr,"EM_JEG: Info: internal error: cannot find a reference in the past to use for concealment\n"); exit(1);
//		}
//		//TComPicYuv* pcYuvRef = pcPic->getPicYuvRec();
//
//		// No border check since it should be covered by the margin
//		// Y
//		for (j=0; j<CUHeight; j++) {
//			//Pel* pYImgCurr = pcYuv->getBufY() + pcYuv->EM_JEG_getCUOffsetY(cu_address) + j*pcYuv->getStride();
//			int margin = pcYuv->getLumaMargin();
//			int marginRef = pcYuvRef->getLumaMargin();
//			Pel* pYImgCurr = pcYuv->getBufY() + margin + pcYuv->EM_JEG_getCUOffsetY(cu_address) + (j+margin)*pcYuv->getStride();
//			Pel* pYImgRef = pcYuvRef->getBufY() + marginRef + pcYuvRef->EM_JEG_getCUOffsetY(cu_address) + (j+marginRef)*pcYuvRef->getStride();
//			for (i=0; i<CUWidth; i++) {
//				*pYImgCurr = *pYImgRef;
//				pYImgCurr++;
//				pYImgRef++;
//			}
//		}
//		// U,V
//		for (j=0; j<CUHeight/2; j++) {
//			int margin = pcYuv->getChromaMargin();
//			int marginRef = pcYuvRef->getChromaMargin();
//			Pel* pUImgCurr = pcYuv->getBufU() + margin + pcYuv->EM_JEG_getCUOffsetC(cu_address) + (j+margin)*pcYuv->getCStride();
//			Pel* pVImgCurr = pcYuv->getBufV() + margin + pcYuv->EM_JEG_getCUOffsetC(cu_address) + (j+margin)*pcYuv->getCStride();
//			Pel* pUImgRef = pcYuvRef->getBufU() + marginRef + pcYuvRef->EM_JEG_getCUOffsetC(cu_address) + (j+marginRef)*pcYuvRef->getCStride();
//			Pel* pVImgRef = pcYuvRef->getBufV() + marginRef + pcYuvRef->EM_JEG_getCUOffsetC(cu_address) + (j+marginRef)*pcYuvRef->getCStride();
//			for (i=0; i<CUWidth/2; i++) {
//				*pUImgCurr = *pUImgRef;
//				pUImgCurr++;
//				pUImgRef++;
//				*pVImgCurr = *pVImgRef;
//				pVImgCurr++;
//				pVImgRef++;
//			}
//		}


	}

}


void TDecEM_JEG::EM_JEG_free(void) {
	EM_JEG_close_file(filedes.loss);
	EM_JEG_close_file(filedes.processed_bitstream);
	free(current_slice_info->cu);
	free(current_slice_info);
	//if (filedes.processed_bitstream) {
	//	// Write the last NALU that has not been written
	//	EM_JEG_write_BitstreamData_if_needed(EM_JEG_nalu);

	//	EM_JEG_close_file(filedes.processed_bitstream);
	//	free(EM_JEG_nalu->buf);
	//	free(EM_JEG_nalu);
	//}
}

void TDecEM_JEG::EM_JEG_dpb_print(void) {
	int i;
	for (i=0; i<EM_JEG_dpb_max; i++) {
		printf("(EM_JEG_dpb: %8p fr %3d valid %d type %d)\n",
			(void *)EM_JEG_dpb[i],
			EM_JEG_dpb_info[i].poc,
			EM_JEG_dpb_info[i].valid,
			EM_JEG_dpb_info[i].type);
	}
	printf("EM_JEG_dpb_size = %d\n", EM_JEG_dpb_size);
}



void TDecEM_JEG::EM_JEG_dpb_print_POCs(void) {
	int i;
	int idx=0;
	for (i=0; i<EM_JEG_dpb_max; i++) {
		if (EM_JEG_dpb_info[i].valid) {
			EM_JEG_debug_dpb_poc[idx] = EM_JEG_dpb_info[i].poc;
			idx++;
		}
	}

	std::qsort( EM_JEG_debug_dpb_poc, idx, sizeof(int), int_cmp);

	printf("EM_JEG_debug_dpb_poc: ");
	for (i=0; i<idx; i++) {
		printf("%d ", EM_JEG_debug_dpb_poc[i]);
	}
	printf("\n");
}


void TDecEM_JEG::EM_JEG_dpb_remove_oldest_picture(void) {
	int i;
	int poc_old = EM_JEG_INT_MAX;
	int idx= -1;
	for (i=0; i<EM_JEG_dpb_max; i++) {
		if (EM_JEG_dpb_info[i].valid>0) {
			if (EM_JEG_dpb_info[i].poc < poc_old) {
				poc_old = EM_JEG_dpb_info[i].poc;
				idx=i;
			}
		}
	}
	if (idx<0) {
		printf("ERROR: EM_JEG_dpb: cannot remove oldest picture: cannot find it\n"); exit(1);
	}
	EM_JEG_dpb_info[idx].poc=EM_JEG_INT_MIN;
	EM_JEG_dpb_info[idx].valid=0;
	EM_JEG_dpb_info[idx].type=-1;
	EM_JEG_dpb_size--;
}


int TDecEM_JEG::EM_JEG_dpb_find_free_element(void) {
	int idx=-1;
	int i;
	for (i=0; i<EM_JEG_dpb_max; i++) {
		if (EM_JEG_dpb_info[i].valid==0) {
			return i;
		}
	}
	return idx;
}

int TDecEM_JEG::EM_JEG_dpb_min_POC(void) {
	int poc_min=EM_JEG_INT_MAX;
	//int idx=-1;
	int i;
	for (i=0; i<EM_JEG_dpb_max; i++) {
		if (EM_JEG_dpb_info[i].valid>0) {
			if (EM_JEG_dpb_info[i].poc < poc_min) {
				poc_min = EM_JEG_dpb_info[i].poc;
				//idx=i;
			}
		}
	}
	return poc_min;
}


// Find the frame with the highest poc < poc and type == B (isB=1) or IoP (isB=0)
// This is useful to conceal the frame
// NB: If poc < EM_JEG_dpb_min_POC() there will be no solution, so change poc to EM_JEG_INT_MAX so that is it like a new poc greater than all is to be concealed
int TDecEM_JEG::EM_JEG_dpb_find_maxframe_and_type(int poc, int isIorP) {
	int poc_max=EM_JEG_INT_MIN;
	int idx=EM_JEG_INT_MIN;
	int i;
	//printf("DEBUG EM_JEG: EM_JEG_dpb_find_maxframe_and_type( poc= %d EM_JEG_dpb_min_POC()=%d )\n", poc, EM_JEG_dpb_min_POC() );
	if (poc < EM_JEG_dpb_min_POC()) {
		poc = EM_JEG_INT_MAX;
	}
	for (i=0; i<EM_JEG_dpb_max; i++) {
		if (EM_JEG_dpb_info[i].valid>0) {
			if (EM_JEG_dpb_info[i].poc < poc) {
				if ( !isIorP || (isIorP &&  (EM_JEG_dpb_info[i].type == I_SLICE || EM_JEG_dpb_info[i].type == P_SLICE)) ) {
					if (EM_JEG_dpb_info[i].poc > poc_max) {
						poc_max = EM_JEG_dpb_info[i].poc;
						idx=i;
					}
				}
			}
		}
	}
	//if (idx!=EM_JEG_INT_MIN) { printf("DEBUG EM_JEG: EM_JEG_dpb_find_maxframe_and_type(poc=%d, isIorP=%d) = %d (poc=%d)\n", poc, isIorP, idx, EM_JEG_dpb_info[idx].poc); } else { printf("DEBUG EM_JEG: EM_JEG_dpb_find_maxframe_and_type(poc=%d, isIorP=%d) = %d (poc= -)\n", poc, isIorP, idx); }
	return idx;
}


// Change the poc of all valid frames in EM_JEG_dpb of the specified amount
void TDecEM_JEG::EM_JEG_dpb_shift_all_POC(int shift_val) {
	int i;
	for (i=0; i<EM_JEG_dpb_max; i++) {
		if (EM_JEG_dpb_info[i].valid>0) {
			EM_JEG_dpb_info[i].poc += shift_val;
		}
	}
}



// ====================================================================================================================
// Public member functions
// ====================================================================================================================

void TDecEM_JEG::EM_JEG_init_frame(int newPoc, int shift) {
        // Check if the new frame has POC ==0 (e.g., a new IDR with poc=0 has arrived)

	// NB: This needs to be done before starting to decode the first NALU of the new frame,
	// since the shift in POC numbers need to happen before that

	// Since it is not easy to detect a new frame from the TDecTop.cpp (see comment describing the bug in that file),
	// the last newPoc is saved,
	// when a new one is encountered a new frame is detected, but the shift is performed only the second time the function is called.

	// NB: This is NEEDED if the dpb size is bigger that the one envisaged by the encoder, since POC may wrap around 
	// (e.g., restart from 0) and create confusion
	// Since a fixed dpb buffer size (16) is used in this code, this is needed (differently from the case of AVC)
	// NB: the dpb is not the one of the decoder, since some frames which would not normally be saved by the decoder are
	// saved for concealment purposes (e.g., sequence of IDR frames)

	if (EM_JEG_doShift == 1) {
		//int EM_JEG_min_POC_in_dpb = EM_JEG_dpb_min_POC();
		if (EM_JEG_dpb_size > 0) {
			//if (dec_picture->poc < EM_JEG_min_POC_in_dpb) 
			//int poc = pcPic->getPOC();
			int poc = newPoc;
			//if (poc == 0 || poc <= EM_JEG_min_POC_in_dpb) {
			if (poc == 0) {
				// Shift all poc values in the buffer so that the new is the highest (e.g., a new IDR with poc=0 has arrived)
				// move of at least 2^ getBitsForPOC()  // log2_max_pic_order_cnt_lsb_minus4  is set in the current SPS
				//int shift = 1 << pcPic->getSlice(0)->getSPS()->getBitsForPOC();
				EM_JEG_dpb_shift_all_POC(-shift); // Must be much lower than 0 (IDR has 0, but some B frames not yet read can be negative)
			}
		}
		EM_JEG_doShift = 0;
	}
	if (newPoc!=EM_JEG_oldPoc) {
		EM_JEG_doShift = 1;
	}
	EM_JEG_oldPoc = newPoc;
}


void TDecEM_JEG::EM_JEG_save_frame(TComPic *pcPic) {
	if (!EM_JEG_dpb) {
		EM_JEG_init_buffers(pcPic);
	}

//	//TEST
//	//char filenameout[] = "test2.yuv";
//	//storedPicYuv->dump(filenameout,1); // 1=append
//	//
//
//	pcPic->getPicYuvRec()->copyToPic(storedPicYuv);
//	storedPicYuvPOC = pcPic->getPicSym()->getSlice(0)->getPOC();
//
//	// Reset after the first call to save_frame, since the whole first frame has been decoded
//	firstPictureFlag = 0;


	if (EM_JEG_dpb_size==EM_JEG_dpb_max) {
		//printf("DEBUG EM_JEG: removing oldest picture\n");
		EM_JEG_dpb_remove_oldest_picture();
	}
	
	int idx = EM_JEG_dpb_find_free_element();
	if (idx<0) {
		printf("ERROR: cannot store picture in EM_JEG_dpb: buffer is full\n"); exit(1);
	}


	// add picture to EM_JEG_dpb in position idx
	TComPicYuv *p = EM_JEG_dpb[idx];

	TComPicYuv *dec_picture = pcPic->getPicYuvRec();

	assert( dec_picture->getLumaMargin() == p->getLumaMargin() );
	assert( dec_picture->getStride() == p->getStride() );
	assert( dec_picture->getChromaMargin() == p->getChromaMargin() );
	assert( dec_picture->getCStride() == p->getCStride() );

	dec_picture->copyToPic(p);	

	//copy_frame(p->imgY,dec_picture->imgY, width, height);
	//copy_frame(p->imgUV[0],dec_picture->imgUV[0], width_cr, height_cr);
	//copy_frame(p->imgUV[1],dec_picture->imgUV[1], width_cr, height_cr);

	//p->poc = dec_picture->poc;
	//p->setPOC( pcPic->getPOC() );  // Not really needed

	//printf("DEBUG EM_JEG: adding picture to my dpb: poc= %d\n", p->poc);
	// Remember to copy also other fields if needed... (e.g., MV for concealment)

	//EM_JEG_dpb_info[idx].poc = dec_picture->poc;
	EM_JEG_dpb_info[idx].poc = pcPic->getPOC();
	EM_JEG_dpb_info[idx].valid = 1;
	//EM_JEG_dpb_info[idx].type = dec_picture->slice_type;
	EM_JEG_dpb_info[idx].type = -1;
	EM_JEG_dpb_size++;

	EM_JEG_frame_num++;

	// Reset CU concealment flags
	m_cpTComEM_JEG->EM_JEG_resetCUConcealed();	

	//EM_JEG_dpb_print();
	//EM_JEG_dpb_print_POCs();
}


void TDecEM_JEG::EM_JEG_add_new_cu(int cu_addr) {
	//printf("DEBUG: current_slice_info->num_cu = %d, cu_addr = %d\n", current_slice_info->num_cu, cu_addr);
	if (current_slice_info->num_cu < MAX_CU_NUM) {
		current_slice_info->cu[current_slice_info->num_cu].cu_addr = cu_addr;  // May not be correct if FMO is enabled
		current_slice_info->cu[current_slice_info->num_cu].type = -1; // Currently unused
		current_slice_info->num_cu++;
	} else {
		printf("Cannot add CU address to current_slice_info: increase MAX_CU_NUM\n"); exit(1);
	}
	return;
}



void TDecEM_JEG::EM_JEG_apply_slice_concealment(TComPic *pcPic) {
	if (filedes.loss) { // Do it only if the loss file is present
		//printf("EM_JEG DEBUG: EM_JEG_apply_slice_concealment() called\n");

		int i;
		int loss_event;

		loss_event = EM_JEG_get_loss_event();

		//EM_JEG_dpb_print_POCs();

		if (loss_event) {
			EM_JEG_nalu_isLost = 1;  // Do not write the NALU in the processed_bitstream

			TComPicYuv* refPic;

			int sliceIdx = pcPic->getCurrSliceIdx();
			//printf("EM_JEG: Info: getCurrSliceIdx() = %d\n", sliceIdx);
			int poc_num__to_be_concealed = pcPic->getSlice(sliceIdx)->getPOC();
			//printf("EM_JEG: Info: POC poc_num__to_be_concealed: %d\n", poc_num__to_be_concealed);

			//TComPicYuv* pcYuv = pcPic->getPicYuvRec();

			// It should work also when POC wraps around, due to EM_JEG_dpb_shift_all_POC() (FIXME: which value to use? ~1000, i.e., 500 frames?)
			int idx = EM_JEG_INT_MIN;

			// I_SLICE P_SLICE B_SLICE
			// pcPic->getSlice(sliceIdx)->getSliceType  // Or any other slice number
			//Char c = ( (pcPic->getSlice(sliceIdx)->isIntra() ) ? 'I' :  ( pcPic->getSlice(sliceIdx)->isInterP() ? 'P' : 'B' ) );

			//if ( img->type==I_SLICE || img->type==P_SLICE ) 
			if ( pcPic->getSlice(sliceIdx)->isIntra()  ||  pcPic->getSlice(sliceIdx)->isInterP() ) {
				idx = EM_JEG_dpb_find_maxframe_and_type(poc_num__to_be_concealed, 0);
			} else {
				// This is a B_SLICE
			
				// OLD COMMENT
				// list1 do NOT contain the last decoded frame. For this, you need to look into the DPB.
				// Only list0 is used for usual buffer management as in MPEG-2 even for backward-predicted frames.
				// list0 and list1 are there for hierarchical frame coding etc.
				// But DPB management is fairly complex (short term references, long term refeences, etc...)
				// Then, after a frame is added with  store_picture_in_dpb()  the buffer is reorganized to make room for new frames
				// by outputting the one not used for reference (in the correct order in case P and B are present...) and then added.
				// Therefore it is very difficult do determine the index at which the frame is stored into the DPB (and it may change at next step).
				// For this reason, a copy is kept locally to simplify the code.
				// Function  EM_JEG_save_ref_frame() implement this
			
				// NEW COMMENT
				// a DPB is needed anyway to ensure correct operation in any conditions (e.g., IIIII etc.)
			
				// Use this code to use the last reference frame, not the last frame as concealment
				//idx = EM_JEG_dpb_find_maxframe_and_type(poc_num__to_be_concealed, 1);
			
				idx = EM_JEG_dpb_find_maxframe_and_type(poc_num__to_be_concealed, 0);
			}

                        if (EM_JEG_dpb_size > 0) {
				refPic = EM_JEG_dpb[idx];
			} else {
				refPic = NULL;  // No pictures decoded yet (to use for concealment)
			}


			//TComPicYuv* pcYuv = pcPic->getPicYuvRec();
			//char filenameout[] = "test.yuv";
			//pcYuv->dump(filenameout,1); // 1=append
        
			//EM_JEG_debug_print=1;
			for (i=0; i<current_slice_info->num_cu; i++) {
				if (loss_event) {
					//EM_JEG_apply_cu_concealment(img, img->dec_picture->imgY, img->dec_picture->imgUV, curr_slice_info->cu[i].cu_addr);
					EM_JEG_apply_cu_concealment(pcPic, refPic, current_slice_info->cu[i].cu_addr);
					//EM_JEG_debug_print=0;
				}
			}
        
			//free(curr_slice_info->cu);
			//free(curr_slice_info);

		}
	}

	// Reset the content of the current slice: prepare for the next one
	current_slice_info->num_cu=0;

	// increment total number of slices
	current_slice_num++;

}


//! \}
