/** \file     TDecEM_JEG.h
    \brief    EM_JEG decoder modification class (header) to consider data loss while decoding frames.

 *
 * Decoder modifications to consider data loss while decoding frames. Source code compatible with HM 12.1
 * Author: Enrico Masala <masala@polito.it>, Politecnico di Torino, Feb 9, 2014
 *
 * Disclaimer:
 * This software is meant to be a prototype only, it is not optimized or fitted to any purposes.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * 
 *
 * Citing this work:
 * We kindly ask people using this code in their research activity
 * to cite the research papers explaining the logic behind this code.
 * More information at: http://media.polito.it/jeg

*/

#ifndef __TDECEM_JEG__
#define __TDECEM_JEG__

#include <list>
#include <vector>

#include <stdlib.h>

#include "TLibCommon/CommonDef.h"
#include "TLibCommon/TComBitStream.h"
#include "TLibCommon/TComList.h"
#include "TLibCommon/TComYuv.h"
#include "TLibCommon/TComPicYuv.h"
#include "TLibCommon/TComPic.h"
//#include "TLibCommon/TComLoopFilter.h"
#include "TLibCommon/TComSampleAdaptiveOffset.h"

//#include "NALread.h"

#include "TLibCommon/TComEM_JEG.h"


#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000


//#define MAX_MB_NUM ((11*4)*(9*4))
// FullHD  1920*1080 = 120*68(67.5)= 8160 MB
// UltraHD  3840*2160 = 240*135 = 32400 MB 
//#define MAX_CU_NUM 32400

#define MAX_NALU_SIZE 10000000

typedef struct EM_JEG_cu_info_s {
	int cu_addr;
	int type;
} EM_JEG_cu_info_t;

typedef struct EM_JEG_slice_info_s {
	EM_JEG_cu_info_t *cu;  // pointer to an array of MB info
	int num_cu;
} EM_JEG_slice_info_t;

typedef struct EM_JEG_fd_s {
	FILE *loss;
	FILE *processed_bitstream;
	FILE *nalu_type;
} EM_JEG_fd_t;

typedef struct EM_JEG_dpb_info_s {
	int poc;
	int valid;
	int type;
} EM_JEG_dpb_info_t;

typedef struct EM_JEG_nalu_s {
	int len;
	uint8_t *buf;
	enum NalUnitType type;
} EM_JEG_nalu_t;


//! \ingroup TLibDecoder
//! \{

// ====================================================================================================================
// Class definition
// ====================================================================================================================



/// GOP decoder class
class TDecEM_JEG
{
private:
  char*                 m_pchEM_JEGLossFile;
  int                   m_pchEM_JEGLossFirst;
  char*                 m_pchEM_JEGProcessedFile;
  char*                 m_pchEM_JEGNALUTypeFile;
  EM_JEG_fd_t filedes;
  //EM_JEG_frame_t storedFrame;
  //TComPicYuv* storedPicYuv;
  //int firstPictureFlag;
  
  int EM_JEG_frame_num;
  int current_slice_num;
  EM_JEG_slice_info_t *current_slice_info;
  //int isCUConcealed[MAX_CU_NUM];
  TComEM_JEG* m_cpTComEM_JEG;
  //int nLossEvents;
  int nSlices;
  TComList<TComPic*>* m_cListPic;

  
  int EM_JEG_debug_print;
  int *EM_JEG_debug_dpb_poc;

  TComPicYuv **EM_JEG_dpb;
  EM_JEG_dpb_info_t *EM_JEG_dpb_info;
  int EM_JEG_dpb_max;
  int EM_JEG_dpb_size;
  int EM_JEG_max_frame_num;
  int EM_JEG_flag_lossfirst;

  EM_JEG_nalu_t *EM_JEG_nalu;
  int EM_JEG_nalu_isLost;

  int EM_JEG_oldPoc;
  int EM_JEG_doShift;

  void EM_JEG_init_buffers(void);

  EM_JEG_slice_info_t*  EM_JEG_get_slice_info(void);
  int EM_JEG_get_loss_event(void);
  void EM_JEG_open_file_loss(void);
  void EM_JEG_open_file_processed_bitstream(void);
  void EM_JEG_open_file_nalutype(void);
  void EM_JEG_close_file(FILE *fp);

  void EM_JEG_free(void);
  void EM_JEG_dpb_print(void);

  // static code need to be in .h , cannot be in .cpp
  static int int_cmp(const void *v1, const void *v2) { return *(const int *)v1 - *(const int *)v2; } 

  void EM_JEG_dpb_print_POCs(void);
  void EM_JEG_dpb_remove_oldest_picture(void);
  int EM_JEG_dpb_find_free_element(void);
  int EM_JEG_dpb_min_POC(void);
  int EM_JEG_dpb_find_maxframe_and_type(int poc, int isIorP);
  void EM_JEG_dpb_shift_all_POC(int shift_val);

  void EM_JEG_init_buffers(TComPic *pcPic);

  void EM_JEG_apply_cu_concealment(TComPic *pcPic, TComPicYuv *refPic, int cu_address);
  //void EM_JEG_allocate_storedPicYuv(TComPic *pcPic);

  void EM_JEG_print_BitstreamData(uint8_t *buf, int len);
  //void EM_JEG_write_BitstreamData_if_needed(EM_JEG_nalu_t *nalu);
  void EM_JEG_private_write_BitstreamData_if_needed(void);
  void EM_JEG_save_Annexb_NALU(uint8_t *data, int len, enum NalUnitType naluType);

public:
  TDecEM_JEG();
  virtual ~TDecEM_JEG();
  
  void init(TComEM_JEG *cpTComEM_JEG);
  void create  ();
  void destroy ();

  void setEM_JEGPicList ( TComList<TComPic*>* picList ) { m_cListPic = picList; };

  void setEM_JEGLossFile ( char *EM_JEGLossFile );
  void setEM_JEGProcessedFile( char *EM_JEGProcessedFile );
  void setEM_JEGLossFirst ( int EM_JEGLossFirst ) { m_pchEM_JEGLossFirst = EM_JEGLossFirst; };
  void setEM_JEGNALUTypeFile( char *EM_JEGNALUTypeFile );
  void setEM_JEGParameterSetManager( ParameterSetManager *parameterSetManager );

  void EM_JEG_init_frame(int newPoc, int shift);
  void EM_JEG_apply_slice_concealment(TComPic *pcPic);
  void EM_JEG_save_frame(TComPic *pcPic);
  void EM_JEG_add_new_cu(int cu_addr);

  int EM_JEG_isCUConcealed(int idx);

  void EM_JEG_saveNALU(vector<uint8_t>& nalUnit, enum NalUnitType naluType);
  void EM_JEG_write_BitstreamData_if_needed(void);

  static const char *EM_JEG_get_NALU_name(enum NalUnitType v) {
    switch (v) {
      case NAL_UNIT_CODED_SLICE_TRAIL_N: return "NAL_UNIT_CODED_SLICE_TRAIL_N"; break;
      case NAL_UNIT_CODED_SLICE_TRAIL_R: return "NAL_UNIT_CODED_SLICE_TRAIL_R"; break;
      
      case NAL_UNIT_CODED_SLICE_TSA_N: return "NAL_UNIT_CODED_SLICE_TSA_N"; break;
      case NAL_UNIT_CODED_SLICE_TSA_R: return "NAL_UNIT_CODED_SLICE_TSA_R"; break;
     
      case NAL_UNIT_CODED_SLICE_STSA_N: return "NAL_UNIT_CODED_SLICE_STSA_N"; break;
      case NAL_UNIT_CODED_SLICE_STSA_R: return "NAL_UNIT_CODED_SLICE_STSA_R"; break;

      case NAL_UNIT_CODED_SLICE_RADL_N: return "NAL_UNIT_CODED_SLICE_RADL_N"; break;
      case NAL_UNIT_CODED_SLICE_RADL_R: return "NAL_UNIT_CODED_SLICE_RADL_R"; break;
    
      case NAL_UNIT_CODED_SLICE_RASL_N: return "NAL_UNIT_CODED_SLICE_RASL_N"; break;
      case NAL_UNIT_CODED_SLICE_RASL_R: return "NAL_UNIT_CODED_SLICE_RASL_R"; break;

      case NAL_UNIT_RESERVED_VCL_N10: return "NAL_UNIT_RESERVED_VCL_N10"; break;
      case NAL_UNIT_RESERVED_VCL_R11: return "NAL_UNIT_RESERVED_VCL_R11"; break;
      case NAL_UNIT_RESERVED_VCL_N12: return "NAL_UNIT_RESERVED_VCL_N12"; break;
      case NAL_UNIT_RESERVED_VCL_R13: return "NAL_UNIT_RESERVED_VCL_R13"; break;
      case NAL_UNIT_RESERVED_VCL_N14: return "NAL_UNIT_RESERVED_VCL_N14"; break;
      case NAL_UNIT_RESERVED_VCL_R15: return "NAL_UNIT_RESERVED_VCL_R15"; break;

      case NAL_UNIT_CODED_SLICE_BLA_W_LP: return "NAL_UNIT_CODED_SLICE_BLA_W_LP"; break;
      case NAL_UNIT_CODED_SLICE_BLA_W_RADL: return "NAL_UNIT_CODED_SLICE_BLA_W_RADL"; break;
      case NAL_UNIT_CODED_SLICE_BLA_N_LP: return "NAL_UNIT_CODED_SLICE_BLA_N_LP"; break;
      case NAL_UNIT_CODED_SLICE_IDR_W_RADL: return "NAL_UNIT_CODED_SLICE_IDR_W_RADL"; break;
      case NAL_UNIT_CODED_SLICE_IDR_N_LP: return "NAL_UNIT_CODED_SLICE_IDR_N_LP"; break;
      case NAL_UNIT_CODED_SLICE_CRA: return "NAL_UNIT_CODED_SLICE_CRA"; break;
      case NAL_UNIT_RESERVED_IRAP_VCL22: return "NAL_UNIT_RESERVED_IRAP_VCL22"; break;
      case NAL_UNIT_RESERVED_IRAP_VCL23: return "NAL_UNIT_RESERVED_IRAP_VCL23"; break;

      case NAL_UNIT_RESERVED_VCL24: return "NAL_UNIT_RESERVED_VCL24"; break;
      case NAL_UNIT_RESERVED_VCL25: return "NAL_UNIT_RESERVED_VCL25"; break;
      case NAL_UNIT_RESERVED_VCL26: return "NAL_UNIT_RESERVED_VCL26"; break;
      case NAL_UNIT_RESERVED_VCL27: return "NAL_UNIT_RESERVED_VCL27"; break;
      case NAL_UNIT_RESERVED_VCL28: return "NAL_UNIT_RESERVED_VCL28"; break;
      case NAL_UNIT_RESERVED_VCL29: return "NAL_UNIT_RESERVED_VCL29"; break;
      case NAL_UNIT_RESERVED_VCL30: return "NAL_UNIT_RESERVED_VCL30"; break;
      case NAL_UNIT_RESERVED_VCL31: return "NAL_UNIT_RESERVED_VCL31"; break;

      case NAL_UNIT_VPS: return "NAL_UNIT_VPS"; break;
      case NAL_UNIT_SPS: return "NAL_UNIT_SPS"; break;
      case NAL_UNIT_PPS: return "NAL_UNIT_PPS"; break;
      case NAL_UNIT_ACCESS_UNIT_DELIMITER: return "NAL_UNIT_ACCESS_UNIT_DELIMITER"; break;
      case NAL_UNIT_EOS: return "NAL_UNIT_EOS"; break;
      case NAL_UNIT_EOB: return "NAL_UNIT_EOB"; break;
      case NAL_UNIT_FILLER_DATA: return "NAL_UNIT_FILLER_DATA"; break;
      case NAL_UNIT_PREFIX_SEI: return "NAL_UNIT_PREFIX_SEI"; break;
      case NAL_UNIT_SUFFIX_SEI: return "NAL_UNIT_SUFFIX_SEI"; break;
      case NAL_UNIT_RESERVED_NVCL41: return "NAL_UNIT_RESERVED_NVCL41"; break;
      case NAL_UNIT_RESERVED_NVCL42: return "NAL_UNIT_RESERVED_NVCL42"; break;
      case NAL_UNIT_RESERVED_NVCL43: return "NAL_UNIT_RESERVED_NVCL43"; break;
      case NAL_UNIT_RESERVED_NVCL44: return "NAL_UNIT_RESERVED_NVCL44"; break;
      case NAL_UNIT_RESERVED_NVCL45: return "NAL_UNIT_RESERVED_NVCL45"; break;
      case NAL_UNIT_RESERVED_NVCL46: return "NAL_UNIT_RESERVED_NVCL46"; break;
      case NAL_UNIT_RESERVED_NVCL47: return "NAL_UNIT_RESERVED_NVCL47"; break;
      case NAL_UNIT_UNSPECIFIED_48: return "NAL_UNIT_UNSPECIFIED_48"; break;
      case NAL_UNIT_UNSPECIFIED_49: return "NAL_UNIT_UNSPECIFIED_49"; break;
      case NAL_UNIT_UNSPECIFIED_50: return "NAL_UNIT_UNSPECIFIED_50"; break;
      case NAL_UNIT_UNSPECIFIED_51: return "NAL_UNIT_UNSPECIFIED_51"; break;
      case NAL_UNIT_UNSPECIFIED_52: return "NAL_UNIT_UNSPECIFIED_52"; break;
      case NAL_UNIT_UNSPECIFIED_53: return "NAL_UNIT_UNSPECIFIED_53"; break;
      case NAL_UNIT_UNSPECIFIED_54: return "NAL_UNIT_UNSPECIFIED_54"; break;
      case NAL_UNIT_UNSPECIFIED_55: return "NAL_UNIT_UNSPECIFIED_55"; break;
      case NAL_UNIT_UNSPECIFIED_56: return "NAL_UNIT_UNSPECIFIED_56"; break;
      case NAL_UNIT_UNSPECIFIED_57: return "NAL_UNIT_UNSPECIFIED_57"; break;
      case NAL_UNIT_UNSPECIFIED_58: return "NAL_UNIT_UNSPECIFIED_58"; break;
      case NAL_UNIT_UNSPECIFIED_59: return "NAL_UNIT_UNSPECIFIED_59"; break;
      case NAL_UNIT_UNSPECIFIED_60: return "NAL_UNIT_UNSPECIFIED_60"; break;
      case NAL_UNIT_UNSPECIFIED_61: return "NAL_UNIT_UNSPECIFIED_61"; break;
      case NAL_UNIT_UNSPECIFIED_62: return "NAL_UNIT_UNSPECIFIED_62"; break;
      case NAL_UNIT_UNSPECIFIED_63: return "NAL_UNIT_UNSPECIFIED_63"; break;
      case NAL_UNIT_INVALID: return "NAL_UNIT_INVALID"; break;
      default: return "UNKNOWN"; break;
    }
  }

};

//! \}

#endif

